#![allow(unused)]

use std::{
    borrow::BorrowMut,
    cell::RefCell,
    ops::DerefMut,
    sync::{mpsc::Receiver as MpscReceiver, Arc, Mutex},
    time::Duration,
};

use crossbeam::{
    channel::{Receiver, Select, Sender},
    select,
};
use serde::Serialize;
// The prelude module re-exports the most common used items from statig.
use statig::prelude::*;
use tauri::{Manager, PhysicalPosition, PhysicalSize, Window};
use underlayer::{Bounds, UnderlayEvent};

use crate::overlay;

pub enum Event {
    State(State),
    Bounds(Bounds),
    Return,
}

#[derive(Clone)]
pub struct Overlay {
    bounds: Bounds,
    window: Window,
    previous: State,
}

pub struct OverlaySettings {
    pub auto_hide: bool,
}
pub type LockedOverlaySettings = Mutex<OverlaySettings>;

fn wrap_underlay_rx(rx: MpscReceiver<UnderlayEvent>) -> Receiver<UnderlayEvent> {
    let (cb_underlay_tx, cb_underlay_rx) = crossbeam::channel::unbounded();
    std::thread::spawn(move || {
        while let Ok(event) = rx.recv() {
            cb_underlay_tx.send(event).ok();
        }
    });
    cb_underlay_rx
}

impl Overlay {
    pub fn new(window: Window, target_title: &str) -> Sender<Event> {
        log::warn!("New overlay object");
        let mut overlay = Self {
            window: window.clone(),
            bounds: Bounds::default(),
            previous: State::hidden(),
        };

        window.manage(Mutex::new(OverlaySettings {
            auto_hide: true,
        }));

        let mut fsm = Overlay::uninitialized_state_machine(overlay).init();

        let underlay_rx = wrap_underlay_rx(underlayer::register(target_title));
        let (ee_tx, ee_rx) = crossbeam::channel::unbounded();
        let (window_event_tx, window_event_rx) = crossbeam::channel::unbounded();

        window.on_window_event(move |event| match event {
            tauri::WindowEvent::Focused(focus) => {
                window_event_tx.send(*focus).ok();
            }
            _ => {}
        });

        std::thread::spawn(move || loop {
            select! {
                recv(underlay_rx) -> msg => {
                    match msg {
                        Ok(event) => {
                                Self::handle_underlay_event(&mut fsm, event);
                        },
                        Err(_) => break,
                    }
                },
                recv(ee_rx) -> msg => {
                    match msg {
                        Ok(event) => {
                                Self::handle_event(&mut fsm, event);
                        },
                        Err(_) => break,
                    }
                },
                recv(window_event_rx) -> msg => {
                    match msg {
                        Ok(focus) => Self::handle_overlay_focused(&mut fsm, focus),
                        Err(_) => break,
                    }
                }
            }
        });

        ee_tx
    }

    fn handle_overlay_focused(fsm: &mut InitializedStateMachine<Overlay>, focus: bool) {
        match fsm.state() {
            State::Visible {} => {
                if focus {
                    fsm.window.set_ignore_cursor_events(true);
                    underlayer::focus_target();
                }
            },
            _ => {}
        }
    }

    fn handle_underlay_event(fsm: &mut InitializedStateMachine<Overlay>, event: UnderlayEvent) {
        let settings = fsm.window.state::<LockedOverlaySettings>();
        match event {
            UnderlayEvent::Attach {
                bounds,
                has_access,
                is_fullscreen,
            } => {
                fsm.handle(&Event::Bounds(bounds));
            }
            UnderlayEvent::MoveResize(bounds) => fsm.handle(&Event::Bounds(bounds)),
            UnderlayEvent::Detach => fsm.handle(&Event::State(State::hidden())),
            UnderlayEvent::Focus => {
                fsm.handle(&Event::State(State::Visible {}))
            }
            UnderlayEvent::Blur => {
                if !fsm.window.is_focused().unwrap()
                    && fsm
                        .window
                        .state::<LockedOverlaySettings>()
                        .lock()
                        .is_ok_and(|s| s.auto_hide)
                {
                    fsm.handle(&Event::State(State::Hidden {}))
                }
            }
            UnderlayEvent::X11FullscreenEvent { is_fullscreen } => {}
        }
    }

    fn handle_event(fsm: &mut InitializedStateMachine<Overlay>, event: Event) {
        fsm.handle(&event);
    }
}

#[state_machine(
    initial = "State::hidden()",
    state(derive(Debug, Clone, PartialEq, Eq, Serialize)),
    superstate(derive(Debug)),
    on_transition = "Self::on_transition"
)]
impl Overlay {
    #[action]
    fn on_enter_visible(&mut self) {
        self.window.set_always_on_top(true);
        self.window.show();
        self.window.set_resizable(true);
        if self.bounds.width > 0 && self.bounds.height > 0 {
            self.window
                .set_position(PhysicalPosition::new(self.bounds.x, self.bounds.y));
            self.window
                .set_size(PhysicalSize::new(self.bounds.width, self.bounds.height));
        }

        log::trace!("on_enter_visible");
    }

    #[action]
    fn on_enter_interactable(&mut self) {
        log::trace!("on_enter_interactable");
        self.window.set_ignore_cursor_events(false);
        self.window.set_focus();
    }

    #[action]
    fn on_enter_hidden(&mut self) {
        log::trace!("on_enter_hidden");
        self.window.hide();
        self.window.set_resizable(false);
    }

    #[action]
    fn on_exit_interactable(&mut self) {
        log::trace!("on_exit_interactable");
        self.window.set_ignore_cursor_events(true);
    }

    #[state(superstate = "super_visible")]
    fn visible(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Handled,
                State::Hidden {} => Transition(State::hidden()),
                State::Interactable {} => Transition(State::interactable()),
            },
            _ => Super,
        }
    }

    #[state(superstate = "super_hidden")]
    fn hidden(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Transition(State::visible()),
                State::Hidden {} => Handled,
                State::Interactable {} => Transition(State::interactable()),
            },
            _ => Super,
        }
    }

    #[state(superstate = "super_interactable")]
    fn interactable(&mut self, event: &Event) -> Response<State> {
        match event {
            Event::State(underlying) => match underlying {
                State::Visible {} => Transition(State::visible()),
                State::Hidden {} => Transition(State::hidden()),
                State::Interactable {} => Handled,
            },
            _ => Super,
        }
    }

    #[superstate(superstate = "common", entry_action = "on_enter_visible")]
    fn super_visible(&mut self, event: &Event) -> Response<State> {
        log::trace!("in super_visible");
        Super
    }

    #[superstate(
        superstate = "super_visible",
        entry_action = "on_enter_interactable",
        exit_action = "on_exit_interactable"
    )]
    fn super_interactable(&mut self, event: &Event) -> Response<State> {
        log::trace!("in super_interactable");
        Super
    }

    #[superstate(
        superstate = "common",
        entry_action = "on_visible",
        entry_action = "on_enter_hidden"
    )]
    fn super_hidden(&mut self, event: &Event) -> Response<State> {
        log::trace!("in super_hidden");
        Super
    }

    #[superstate]
    fn common(&mut self, event: &Event) -> Response<State> {
        log::trace!("in common");
        match event {
            Event::Bounds(bounds) => {
                self.bounds = *bounds;
                if bounds.width > 0 && bounds.height > 0 {
                    self.window
                        .set_position(PhysicalPosition::new(bounds.x, bounds.y));
                    self.window
                        .set_size(PhysicalSize::new(bounds.width, bounds.height));
                }
                Super
            }
            Event::Return => {
                log::trace!("Returning to: {:?}", self.previous);
                Transition(self.previous.clone())
            }
            _ => Super,
        }
    }

    fn on_transition(&mut self, src: &State, dest: &State) {
        log::trace!("from {src:?} to {dest:?}");
        if src != dest {
            self.previous = src.clone();
        }
        self.window.emit("OverlayStateChange", dest);
    }
}
